<?php
/**
 * Created by PhpStorm.
 * User: crose
 * Date: 4/25/16
 * Time: 10:39 PM
 */

namespace IMATHUZH\Qfq\Core;


use IMATHUZH\Qfq\Core\Helper\HelperFile;
use IMATHUZH\Qfq\Core\Helper\Logger;
use IMATHUZH\Qfq\Core\Helper\Path;
use IMATHUZH\Qfq\Core\Helper\Sanitize;
use IMATHUZH\Qfq\Core\Helper\Support;
use IMATHUZH\Qfq\Core\Store\Session;
use IMATHUZH\Qfq\Core\Store\Store;

/**
 * Class File
 * @package qfq
 */
class File {

    private $uploadErrMsg = array();
    public $imageUploadFilePath = null;

    /**
     * @var Store
     */
    private $store = null;

    /**
     * @var Session
     */
    private $session = null;

    private $qfqLogPathFilenameAbsolute = '';

    /**
     * @param bool|false $phpUnit
     * @throws \CodeException
     * @throws \UserFormException
     * @throws \UserReportException
     */
    public function __construct($phpUnit = false) {

        #TODO: rewrite $phpUnit to: "if (!defined('PHPUNIT_QFQ')) {...}"
        $this->session = Session::getInstance($phpUnit);

        $this->store = Store::getInstance('', $phpUnit);
        $this->qfqLogPathFilenameAbsolute = Path::absoluteQfqLogFile();

        $this->uploadErrMsg = [
            UPLOAD_ERR_INI_SIZE => "The uploaded file exceeds the upload_max_filesize directive in php.ini",
            UPLOAD_ERR_FORM_SIZE => "The uploaded file exceeds the MAX_FILE_SIZE directive that was specified in the HTML form",
            UPLOAD_ERR_PARTIAL => "The uploaded file was only partially uploaded",
            UPLOAD_ERR_NO_FILE => "No file was uploaded",
            UPLOAD_ERR_NO_TMP_DIR => "Missing a temporary folder",
            UPLOAD_ERR_CANT_WRITE => "Failed to write file to disk",
            UPLOAD_ERR_EXTENSION => "File upload stopped by extension",
        ];
    }

    /**
     * @throws \CodeException
     * @throws \UserFormException
     */
    public function process() {

        $action = $this->store->getVar(FILE_ACTION, STORE_CLIENT, SANITIZE_ALLOW_ALNUMX);
        $sipUpload = $this->store->getVar(SIP_SIP, STORE_SIP);
        if ($sipUpload === false && $action != FILE_ACTION_IMAGE_UPLOAD) {

            if (empty($_FILES)) {
                throw new \UserFormException('Missing $_FILES[] - the whole request seems broken', ERROR_UPLOAD_FILES_BROKEN);
            }

            throw new \UserFormException('SIP invalid: ' . $sipUpload, ERROR_SIP_INVALID);
        }

        // Throws an exception if content is too big - if content is bigger than 'post_max_size', the POST is lost together with the PHP Upload error message.
        $size = $_SERVER['CONTENT_LENGTH'] ?? 0;

        if ($action != FILE_ACTION_IMAGE_UPLOAD) {
            $this->checkMaxFileSize($size);
        }

        $statusUpload = $this->store->getVar($sipUpload, STORE_EXTRA, SANITIZE_ALLOW_ALL);
        if ($statusUpload === false) {
            $statusUpload = array();
        }


        switch ($action) {
            case FILE_ACTION_UPLOAD:
                $this->doUpload($sipUpload, $statusUpload);
                break;
            case FILE_ACTION_DELETE:
                $this->doDelete($sipUpload, $statusUpload);
                break;
            case FILE_ACTION_IMAGE_UPLOAD:
                $this->imageUploadFilePath = $this->doImageUpload();
                break;
            default:
                throw new \UserFormException("Unknown FILE_ACTION: $action", ERROR_UPLOAD_UNKNOWN_ACTION);
        }
    }

    /**
     * Checks the max file size, defined per FormElement.
     * Only possible if there is a valid SIP Store.
     *
     * @param $size
     * @throws \CodeException
     * @throws \UserFormException
     */
    private function checkMaxFileSize($size) {

        // Checked and set during form build.
        $maxFileSize = $this->store->getVar(FE_FILE_MAX_FILE_SIZE, STORE_SIP . STORE_ZERO);

        if ($size >= $maxFileSize) {
            throw new \UserFormException('File too big. Max allowed size: ' . $maxFileSize . ' Bytes', ERROR_UPLOAD_TOO_BIG);
        }
    }

    /**
     * doUpload
     *
     * @param string $sipUpload
     * @param array $statusUpload
     *
     * @throws \CodeException
     * @throws \UserFormException
     */
    private function doUpload($sipUpload, array $statusUpload) {

        // New upload
        $newArr = reset($_FILES);
        // Merge new upload date to existing status information
        $statusUpload = array_merge($statusUpload, $newArr);
        HelperFile::fixMimeTypeHeif($statusUpload['tmp_name'], $statusUpload['type']);

        if ($statusUpload[FILES_ERROR] !== UPLOAD_ERR_OK) {
            throw new \UserFormException(
                json_encode([ERROR_MESSAGE_TO_USER => 'Upload: Error', ERROR_MESSAGE_TO_DEVELOPER => $this->uploadErrMsg[$newArr[FILES_ERROR]]]),
                ERROR_UPLOAD);
        }

        Logger::logMessageWithPrefix(UPLOAD_LOG_PREFIX . ': File under ' . $statusUpload['tmp_name'], $this->qfqLogPathFilenameAbsolute);

        $this->checkMaxFileSize($statusUpload['size']);

        $accept = $this->store->getVar(FE_FILE_MIME_TYPE_ACCEPT, STORE_SIP);
        if ($accept != '' && !HelperFile::checkFileType($statusUpload['tmp_name'], $statusUpload['name'], $accept)) {
            throw new \UserFormException('Filetype not allowed. Allowed: ' . $accept, ERROR_UPLOAD_FILE_TYPE);
        }

        // rename uploaded file: ?.cached
        $filenameCached = Support::extendFilename($statusUpload[FILES_TMP_NAME], UPLOAD_CACHED);

        // Remember: PHP['upload_tmp_dir']='' means '/tmp' AND upload process is CHROOT to /tmp/systemd-private-...-apache2.service-../
        error_clear_last();
        if (!move_uploaded_file($newArr[FILES_TMP_NAME], $filenameCached)) {
            $msg = error_get_last();
            throw new \UserFormException(
                json_encode([ERROR_MESSAGE_TO_USER => 'Upload: Error', ERROR_MESSAGE_TO_DEVELOPER => $msg]),
                ERROR_UPLOAD_FILE_TYPE);
        }

        $this->store->setVar($sipUpload, $statusUpload, STORE_EXTRA);
    }

    /**
     * @param $sipUpload
     * @param $statusUpload
     *
     * @throws \CodeException
     * @throws \UserFormException
     * @internal param string $keyStoreExtra
     */
    private function doDelete($sipUpload, $statusUpload) {

        if (isset($statusUpload[FILES_TMP_NAME]) && $statusUpload[FILES_TMP_NAME] != '') {
            $file = Support::extendFilename($statusUpload[FILES_TMP_NAME], UPLOAD_CACHED);
            if (file_exists($file)) {
                HelperFile::unlink($file, $this->qfqLogPathFilenameAbsolute);
            }
            $statusUpload[FILES_TMP_NAME] = '';
        }

        $statusUpload[FILES_FLAG_DELETE] = '1';
        $this->store->setVar($sipUpload, $statusUpload, STORE_EXTRA);
    }

    /**
     * Get tinyMce image upload and store it with unique filename in system. Return new filepath.
     *
     * @throws \CodeException
     * @throws \UserFormException
     */
    private function doImageUpload() {
        reset($_FILES);
        $temp = current($_FILES);

        $httpOrigin = Store::getVar(SYSTEM_HTTP_ORIGIN, STORE_SYSTEM, SANITIZE_ALLOW_ALL);

        if (!is_uploaded_file($temp[FILES_TMP_NAME])) {
            // Notify editor that the upload failed
            header(HTTP_500_SERVER_ERROR);
            Logger::logMessageWithPrefix('Internal error. Uploaded file not found.', Path::absoluteQfqLogFile());
            return 0;
        }

        // same-origin requests won't set an origin. If the origin is set, it must be valid.
        if ($httpOrigin === $_SERVER['HTTP_ORIGIN']) {
            header('Access-Control-Allow-Origin: ' . $httpOrigin);
        } else {
            header(HTTP_403_ORIGIN_DENIED);
            Logger::logMessageWithPrefix('Origin denied. From server given origin doesnt match with baseUrl.', Path::absoluteQfqLogFile());
            return 0;
        }

        /*
          If your script needs to receive cookies, set images_upload_credentials : true in
          the configuration and enable the following two headers.
        */
        // header('Access-Control-Allow-Credentials: true');
        // header('P3P: CP="There is no P3P policy."');

        // Sanitize input
        // Disallow 'none alphanumeric'. Allow dot or underscore and conditionally '/'.
        $temp[FILES_NAME] = Sanitize::safeFilename($temp[FILES_NAME]);

        // Verify extension
        // Split between 'Media Type' and 'Media Subtype'
        $fileMimeTypeSplitted = explode('/', $temp[FILES_TYPE], 2);
        if ($fileMimeTypeSplitted[0] !== NAME_IMAGE) {
            header(HTTP_400_INVALID_EXTENSION);
            Logger::logMessageWithPrefix('Mime type error. Invalid picture type.', Path::absoluteQfqLogFile());
            return 0;
        }

        // Base folder path should be with user input changeable -> take over sip given parameters from formElement
        $imageUploadDirUser = Store::getVar(FE_EDITOR_FILE_UPLOAD_PATH, STORE_SIP, SANITIZE_ALLOW_ALL);
        $imageUploadDirUser = base64_decode($imageUploadDirUser);

        // Get upload path from config
        $imageUploadDir = Store::getVar(SYSTEM_IMAGE_UPLOAD_DIR, STORE_SYSTEM, SANITIZE_ALLOW_ALL);

        //overwrite imageUploadDir with user given path if exists
        if ($imageUploadDirUser != '') {
            $imageUploadDir = $imageUploadDirUser;
        }

        $imageUploadDirAbsolute = Path::absoluteApp($imageUploadDir);

        // Name should be unique, add to end of name number if exists to not overwrite it
        $changedFileName = HelperFile::getUniqueFileName($imageUploadDirAbsolute, $temp[FILES_NAME]);
        $neededSlash = '';
        if (substr($imageUploadDirAbsolute, -1) !== '/') {
            $neededSlash = '/';
        }
        $fileToWrite = $imageUploadDirAbsolute . $neededSlash . $changedFileName;
        HelperFile::mkDirParent($fileToWrite);
        move_uploaded_file($temp[FILES_TMP_NAME], $fileToWrite);

        // Respond to the successful upload with JSON.
        return $imageUploadDir . $neededSlash . $changedFileName;
    }
}